#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define MAXLINE 80
#define BUFSZ	8

void do_remotetmp()
{
	char login[MAXLINE], buffer[MAXLINE];
	char c, *p;
	int fd, written, toread;

	if (chdir("tmp") < 0) {
		perror("tmp");
		return;
	}

	printf("user: ");
	if (fgets(login, MAXLINE, stdin) == NULL)
		return;
	if (p = index(login, '\n'))
		*p = '\0';

	fd = open(login, O_RDWR | O_CREAT | O_EXCL, 0600);
	if (fd < 0) {
		if (errno == EEXIST) {
			fprintf(stderr, "already logged in\n");
		} else {
			perror(login);
		}
		return;
	}

	printf("OK\n");
	written = 0;
	while (fgets(buffer, MAXLINE, stdin)) {
		if (p = index(buffer, '\n'))
			*p = '\0';
		switch (buffer[0]) {
		case 'w':
			for (p = buffer + 1; *p; p++) {
				if (write(fd, p, 1) != 1) {
					perror("write");
					break;
				}
				written++;
			}
			break;
		case 'r':
			if (lseek(fd, 0, SEEK_SET) < 0) {
				perror("seek");
				break;
			}
			for (toread = written; toread > 0; toread--) {
				switch (read(fd, &c, 1)) {
				case 1:
					putchar(c);
					break;
				case 0:
					break;
				default:
					perror("read");
					break;
				}
			}
			putchar('\n');
		case '\0':
			break;
		case 'q':
			goto out;
		default:
			fprintf(stderr, "unknown command\n");
			break;
		}
	}
out:
	close(fd);
	if (unlink(login) < 0)
		perror(login);
}

void do_privtmp()
{
	char secret[MAXLINE];
	char c;
	FILE *f, *s;

	snprintf(secret, MAXLINE, "sec.%c", (rand() % 5) + 'A');
	s = fopen(secret, "w");
	if (s == NULL)
		return;
	f = fopen("flag.txt", "r");
	if (f == NULL)
		return;
	while ( (c = fgetc(f)) != EOF )
		fputc(c, s);
	fclose(f);
	printf("done\n");
	
	/* TODO: implement protected manipulation of secret */

	fclose(s);
	unlink(secret);
}

void child()
{
	char buffer[MAXLINE], *p;

	printf("Secure remote tmp server v1.0\n");

	if (fgets(buffer, MAXLINE, stdin) == NULL)
		return;
	if (p = index(buffer, '\n'))
		*p = '\0';
	switch (buffer[0]) {
	case 'r':
		do_remotetmp();
		break;
	case 'p':
		do_privtmp();
		break;
	case '\0':
		break;
	default:
		fprintf(stderr, "unknown command\n");
		break;
	}
}

void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	srand(time(NULL));

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}
